import simpy

from starling_sim.basemodel.agent.stations.station import Station
from starling_sim.basemodel.agent.requests import StationRequest


class VehicleSharingStation(Station):
    """
    Station dedicated to a shared vehicle system
    """

    SCHEMA = {
        "properties": {
            "capacity": {
                "type": "integer",
                "title": "Station capacity",
                "description": "Maximum storing capacity of the station",
                "minimum": 0
            }
        },
        "required": ["capacity"]
    }

    def __init__(self, simulation_model, agent_id, origin, capacity, **kwargs):

        Station.__init__(self, simulation_model, agent_id, origin, **kwargs)

        self.capacity = capacity
        self.initial_stock = 0
        self.store = simpy.Store(self.sim.scheduler.env, capacity=capacity)

    def __str__(self):

        return "[id={}, position={}, capacity={}, initialStock={}]" \
            .format(self.id, self.position, self.capacity, self.initial_stock)

    def get_store(self):
        """
        Returns the station's store
        :return:
        """
        return self.store

    def nb_products(self):
        """
        Returns the number of products in store
        :return:
        """
        return len(self.store.items)

    def get_from_store(self, agent):
        """
        Requests a product from the station's store
        :param agent: requesting agent
        :return: StationRequest object
        """

        request = StationRequest(agent, self.sim.scheduler.now(),
                                 self, StationRequest.GET_REQUEST)

        request.set_request_event(self.store.get())

        return request

    def check(product):
        """
        Tests the type of the product according to the type of store
        This method must be extended for specific stations
        :param product: product to be tester
        :return:
        """

        return True

    check = staticmethod(check)

    def return_to_store(self, agent, product):
        """
        Tries to return a product in the station's store
        :param agent: agent returning product
        :param product: returned product
        :return: StationRequest object
        """

        if self.check(product):

            request = StationRequest(agent, self.sim.scheduler.now(),
                                     self, StationRequest.PUT_REQUEST)

            request.set_request_event(self.store.put(product))

            return request
        else:
            self.log_message("Received wrong type of product ({}) from {}"
                             .format(type(product), agent.id))
            return None

    def set_products(self, products):
        """
        Sets the products contained in the store
        :param products: list of products, type is not checked
        :return:
        """

        self.store.items = products



from starling_sim.basemodel.input.dynamic_input import DynamicInput


class Input(DynamicInput):

    def new_agent_input(self, feature):

        new_agent = super().new_agent_input(feature)

        input_dict = feature["properties"]

        if input_dict["agent_type"] == "vehicle":
            if "station" in input_dict:
                vehicle_station = self.sim.agentPopulation["station"][input_dict["station"]]
            else:
                vehicle_station = self.sim.environment.get_agent_at(new_agent.position,
                                                                    self.sim.agentPopulation["station"].values())
                if vehicle_station is None:
                    self.log_message("No station found for vehicle {}".format(new_agent), 40)
                    exit(1)
            vehicle_station.store.items.append(new_agent)
            vehicle_station.initial_stock += 1

    def pre_process_input_dict(self, input_dict):

        # TODO : use a default fleet and stations name when there is no operator

        if input_dict["agent_type"] == "station":

            operator = self.sim.agentPopulation.population["operator"][input_dict["operator_id"]]

            # add station population
            input_dict["population"] = operator.stations_dict_name

        elif input_dict["agent_type"] == "user":

            super().pre_process_input_dict(input_dict)

            if "origin_station" in input_dict:
                user_station = self.sim.agentPopulation["station"][input_dict["origin_station"]]
                input_dict["origin"] = user_station.position

            if "destination_station" in input_dict:
                user_station = self.sim.agentPopulation["station"][input_dict["destination_station"]]
                input_dict["destination"] = user_station.position

        elif input_dict["agent_type"] == "staff":

            # add origin at depot
            # TODO : adapt for multiple depots
            operator = self.sim.agentPopulation.population["operator"][input_dict["operator_id"]]
            input_dict["origin"] = list(operator.depotPoints.values())[0].position

        elif input_dict["agent_type"] == "vehicle":
            operator = self.sim.agentPopulation.population["operator"][input_dict["operator_id"]]
            vehicle_station = operator.stations[input_dict["station"]]
            input_dict["origin"] = vehicle_station.position



"""
Starling setup script.

See license in LICENSE.txt.
"""

import setuptools
import os
from starling_sim.version import __version__

# short description of the project
DESC = "Agent-based framework for mobility simulation"

# get long description from README.md
# with open("README.md", "r") as fh:
#     LONG_DESC = fh.read()
LONG_DESC = "Long description of the Starling project (TODO)"

# list of classifiers from the PyPI classifiers trove
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Scientific/Engineering :: GIS",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "License :: CeCILL-B Free Software License Agreement (CECILL-B)"
]

# only specify install_requires if not in RTD environment
if os.getenv("READTHEDOCS") == "True":
    INSTALL_REQUIRES = []
else:
    with open("requirements.txt") as f:
        INSTALL_REQUIRES = [line.strip() for line in f.readlines()]

# call setup
setuptools.setup(
    name="starling-sim",
    version=__version__,
    license="CECILL-B",
    author="Tellae",
    author_email="starling@tellae.fr",
    description=DESC,
    long_description=LONG_DESC,
    long_description_content_type="text/markdown",
    url="https://github.com/tellae/starling",
    packages=setuptools.find_packages() + ["starling_sim/schemas"],
    classifiers=CLASSIFIERS,
    python_requires='>=3.6',
    install_requires=INSTALL_REQUIRES,
    include_package_data=True
)



import logging
import osmnx as ox
import networkx as nx
import numpy as np

from starling_sim.basemodel.topology.topology import Topology
from starling_sim.utils.paths import graph_speeds_folder
from starling_sim.utils.utils import json_load, osm_graph_from_file


class OSMNetwork(Topology):
    """
    Open Street Map network topology

    Stores the network as a networkx graph
    """

    IMPORT_ERROR = "Cannot import OSM network : "

    def __init__(self, transport_mode, network_file=None, speed_file=None, graph=None, store_paths=None):
        """
        Create the topology structure, without initializing the network

        :param transport_mode: type of the given network
        :param network_file: graphml file for network import
        :param speed_file: json file containing the link speeds
        :param graph: graph object for direct initialisation
        :param store_paths: boolean indicating if shortest paths should be stored
        """
        super().__init__()

        self.mode = transport_mode
        self.network_file = network_file
        self.speed_file = speed_file

        self.graph = graph
        self.speeds = None

        # paths storing
        if store_paths is None:
            self.store_paths = False
        else:
            self.store_paths = store_paths
        self.paths = {"time": {}, "length": {}}

        # TODO : replace with a decorator ?
        self.shortest_path_count = 0

    def setup(self):
        """
        Prepare the class for the simulation

        Import the OSM graph, set the speeds
        """

        if self.graph is None:

            if self.network_file is None:
                logging.error("No network file provided for topology initialisation")
                raise ValueError("Network file provided is {}".format(self.network_file))

            else:
                logging.debug("Importing OSM graph for mode '{}' from file {}".format(self.mode, self.network_file))
                self.graph = osm_graph_from_file(self.network_file)

        if self.speed_file is None:
            logging.error("No speed file provided for topology initialisation")
            raise ValueError("Speed file provided is {}".format(self.network_file))
        else:
            logging.debug("Importing graph speeds for mode '{}' from file {}".format(self.mode, self.speed_file))
            speeds = json_load(graph_speeds_folder() + self.speed_file)
            self.update_speeds(speeds)

    def update(self):
        """
        Update of the osm network during the simulation
        """

        # for now we don't update the network
        pass

    def get_positions(self):
        """
        Return the list of all topology positions

        :return: list of all graph nodes
        """

        return self.graph.nodes

    # group of static methods for the network management using osmnx and networkx

    def update_speeds(self, speeds):
        """
        Set the speed values according to the given dictionary, and then the time values

        :param speeds: speed dict {<link_type>: {speed: <speed_value>, ...}
        """

        for u, v, d in self.graph.edges(data=True):

            if isinstance(d["highway"], list):
                d["highway"] = d["highway"][0]

            # differentiate speeds among the link types
            if d["highway"] in speeds:
                speed = speeds[d["highway"]]["speed"]
            else:
                speed = speeds["other"]["speed"]

            d["speed"] = speed

            # round the length values
            d["length"] = round(d["length"])

            # time is valued in seconds
            d["time"] = round(3600 * ((d['length']/1000) / speed))

        self.speeds = speeds

    def shortest_path(self, origin, destination, dimension):
        """
        Compute the shortest path between origin and destination

        Raises an exception if origin or destination is None,
        instead of computing all shortest path (this is what nx would do)

        :param origin: origin node
        :param destination: destination node
        :param dimension: type of 'length' to consider (distance, time..)
        :return: shortest path : list of nodes, origin and destination included
        """

        if origin is None or destination is None:
            logging.error("Origin or destination provided for shortest path is None")
            raise ValueError((origin, destination))

        path, _ = self.dijkstra_shortest_path_and_length(origin, destination, dimension)

        return path

    def shortest_path_length(self, origin, destination, dimension):
        """
        Compute the length of the shortest path between origin and destination

        :param origin:
        :param destination:
        :param dimension:
        :return: length of the shortest path
        """

        if origin is None or destination is None:
            logging.error("Origin or destination provided for shortest path is None")
            raise ValueError((origin, destination))

        _, length = self.dijkstra_shortest_path_and_length(origin, destination, dimension)

        return length

    def dijkstra_shortest_path_and_length(self, origin, destination, dimension):
        """
        Compute a shortest path and its length using Dijkstra's algorithm

        :param origin: source node
        :param destination: target node
        :param dimension: weight used for computing the path length
        :return: tuple (path, path_length)
        """

        od = (origin, destination)

        if self.store_paths and od in self.paths[dimension]:
            return self.paths[dimension][od][0], self.paths[dimension][od][1]
        else:

            self.shortest_path_count += 1

            length, path = nx.single_source_dijkstra(self.graph, origin,
                                                     target=destination, weight=dimension)

            if self.store_paths:
                self.paths[dimension][od] = path, length

            return path, length

    def position_localisation(self, position):

        return [self.graph.nodes[position]["y"], self.graph.nodes[position]["x"]]

    def nearest_position(self, localisation):

        position = ox.get_nearest_node(self.graph, (float(localisation[0]), float(localisation[1])))
        return position

    def localisations_nearest_nodes(self, x_coordinates, y_coordinates, method="balltree"):

        # convert coordinate lists to np.array
        x_array = np.array(x_coordinates, dtype=np.float32)
        y_array = np.array(y_coordinates, dtype=np.float32)

        # we use osmnx get_nearest_nodes function, with method='balltree' [uses scikit_learn]
        return ox.get_nearest_nodes(self.graph, x_array, y_array, method=method)

    # TODO : add a check ?
    def get_edge_data(self, node1, node2, data):

        d = self.graph.get_edge_data(node1, node2)

        return min(attr.get(data) for attr in d.values())



class Topology:
    """
    This class describes an type of environment for the simulation

    It should be extended for concrete use in the simulation
    """

    def __init__(self):
        """
        The constructor should not instantiate the topology data structure,
        which is done in the setup method
        """

        # transport mode described by this topology
        self.mode = None

    def setup(self):
        """
        Prepare the topology for the simulation run

        :return:
        """

    # TODO : doubles with udpate_speeds ?
    def update(self):
        """
        Update the topology state

        :return:
        """

    def get_positions(self):
        """
        Return the list of all topology positions

        :return: list of positions
        """

    # TODO : structure of speeds, and doc
    def update_speeds(self, speeds):
        """

        :param speeds:
        :return:
        """

    # TODO : unify all shortest path computations under the same method ?
    def shortest_path(self, origin, destination, dimension):
        """
        Compute the shortest path between origin and destination

        :param origin: origin position
        :param destination: destination position
        :param dimension: type of length to consider (distance, time, etc)
        :return: shortest path : list of positions, origin and destination included
        """

    # TODO : other shortest path methods ?

    def position_localisation(self, position):
        """
        Return the localisation (lat, lon) of the position

        :param position: position in the topology
        :return: tuple (lat, lon)
        """

    def nearest_position(self, localisation):
        """
        Return the nearest position to given localisation (lat, lon)

        :param localisation: (lat, lon) tuple
        :return: position
        """

    def localisations_nearest_nodes(self, x_coordinates, y_coordinates, method=None):
        """
        Return the graph nodes nearest to a list of points.

        This method should allow a faster computation than multiple nearest_position calls.

        :param x_coordinates: list of X coordinates of the localisations (lon in GPS)
        :param y_coordinates: list of Y coordinates of the localisations (lat in GPS)
        :param method: name of the method used for the computation
        :return: list of nearest nodes
        """

    def get_edge_data(self, node1, node2, data):
        """
        Return the corresponding edge information

        :param node1: node of the topology
        :param node2: node of the topology
        :param data: requested attribute
        :return: data information of the edge
        """



